#!/usr/bin/env python

# pylint: disable=redefined-outer-name

"""xdotool tests."""

import logging
import os

from random import randint

import pytest

from xdotoolp import (
    get_desktop,
    get_desktop_for_window,
    get_num_desktops,
    getactivewindow,
    getwindowgeometry,
    getwindowname,
    getwindowpid,
    guess_window,
    launch_command,
    search,
    set_desktop_for_window,
    windowactivate,
    windowmove,
    windowsize,
)

from .testutils import kill_all_xclock_instances, QUASI_DETERMINISTIC_DELAY


LOGGER = logging.getLogger(__name__)


@pytest.mark.skipif("TRAVIS" in os.environ, reason="Doesn't work with xvfb.")
def test_get_desktop():
    """Tests that the current desktop can be retrieved."""
    assert get_desktop()


@pytest.mark.skipif("TRAVIS" in os.environ, reason="Doesn't work with xvfb.")
@pytest.mark.xclock
def test_get_desktop_for_window():
    """Tests that desktop assigned to a window can be retrieved."""
    try:
        window_metadata = launch_command(args=["xclock"])
        window_id = guess_window(title_hint="^xclock$", window_metadata=window_metadata)
        assert window_id

        assert get_desktop_for_window(window_id=window_id)
    finally:
        kill_all_xclock_instances()


@pytest.mark.skipif("TRAVIS" in os.environ, reason="Doesn't work with xvfb.")
def test_get_num_desktop():
    """Tests that the number of desktops can be retrieved."""
    assert get_num_desktops()


@pytest.mark.skipif("TRAVIS" in os.environ, reason="Doesn't work with xvfb.")
@pytest.mark.xclock
def test_getactivewindow():
    """Tests that the active window can be retrieved.."""
    try:
        window_metadata = launch_command(args=["xclock"])
        window_id = guess_window(title_hint="^xclock$", window_metadata=window_metadata)
        assert window_id

        windowactivate(sync=True, window_id=window_id)
        assert getactivewindow()

    finally:
        kill_all_xclock_instances()


@pytest.mark.xclock
def test_getwindowgeometry():
    """Tests that geometry can be retrieved for a window."""
    try:
        window_metadata = launch_command(args=["xclock"])
        window_id = guess_window(title_hint="^xclock$", window_metadata=window_metadata)
        assert window_id

        match = getwindowgeometry(window_id=window_id)
        assert match.group("geometry")
        assert match.group("position")
        assert match.group("screen")
        assert match.group("window")
    finally:
        kill_all_xclock_instances()


@pytest.mark.xclock
def test_getwindowname():
    """Tests that the name can be retrieved for a window."""
    try:
        window_metadata = launch_command(args=["xclock"])
        window_id = guess_window(title_hint="^xclock$", window_metadata=window_metadata)
        assert window_id

        assert getwindowname(window_id=window_id)
    finally:
        kill_all_xclock_instances()


@pytest.mark.xclock
def test_getwindowpid():
    """Tests that a pid can be retrieved for a window."""
    try:
        window_metadata = launch_command(args=["xclock"])
        window_id = guess_window(title_hint="^xclock$", window_metadata=window_metadata)
        assert window_id

        assert getwindowpid(window_id=window_id)
    finally:
        kill_all_xclock_instances()


@pytest.mark.xclock
def test_search():
    """Tests that a search can be performed to find a window."""
    try:
        window_metadata = launch_command(args=["xclock"])
        window_id = guess_window(title_hint="^xclock$", window_metadata=window_metadata)
        assert window_id

        assert search(
            "--class",
            "--classname",
            "--maxdepth=2",
            "--name",
            "--onlyvisible",
            "^xclock$",
        )
    finally:
        kill_all_xclock_instances()


# TODO: How to test selectwindow()?!?


@pytest.mark.skipif("TRAVIS" in os.environ, reason="Doesn't work with xvfb.")
@pytest.mark.xclock
def test_set_desktop_for_window():
    """Tests that a desktop can be assigned to a window."""
    num_desktops = get_num_desktops()
    if int(num_desktops) < 2:
        pytest.skip(f"Number of desktops {num_desktops} is less than required value: 2")
        return

    try:

        def set_desktop(*, desktop: str):
            set_desktop_for_window(desktop=desktop, window_id=window_id)
            from time import sleep

            sleep(
                QUASI_DETERMINISTIC_DELAY
            )  # There is no "sync" for set_desktop_for_window =/ ...
            assert get_desktop_for_window(window_id=window_id) == desktop

        window_metadata = launch_command(args=["xclock"])
        window_id = guess_window(title_hint="^xclock$", window_metadata=window_metadata)
        assert window_id

        # After the next command window_id will be on desktop=0, but was it before (by default)?
        set_desktop(desktop="0")

        # Since we know it changed, we can be confident that xdotool moved it ...
        set_desktop(desktop="1")
    finally:
        kill_all_xclock_instances()


@pytest.mark.skipif("TRAVIS" in os.environ, reason="Doesn't work with xvfb.")
@pytest.mark.xclock
def test_windowactivate():
    """Tests that a window can be activated."""
    try:

        def make_active(*, window_id: str):
            windowactivate(sync=True, window_id=window_id)
            assert getactivewindow() == window_id

        window_metadata0 = launch_command(args=["xclock"])
        window_id0 = guess_window(
            title_hint="^xclock$", window_metadata=window_metadata0
        )
        assert window_id0

        # After the next command window_id0 will be active, but was it before (by default)?
        make_active(window_id=window_id0)

        window_metadata1 = launch_command(args=["xclock"])
        window_id1 = guess_window(
            title_hint="^xclock$", window_metadata=window_metadata1
        )
        assert window_id1

        # After the next command window_id1 will be active, but was it before (by default)?
        make_active(window_id=window_id1)

        # Since no new windows were opened, we can be confident that it's xdotool changing focus,
        # and not the display manager ...
        make_active(window_id=window_id0)

        # One more time, for good measure ...
        make_active(window_id=window_id1)
    finally:
        kill_all_xclock_instances()


@pytest.mark.xclock
def test_windowmove():
    """Tests that a window can be moved."""
    try:
        window_metadata = launch_command(args=["xclock"])
        window_id = guess_window(title_hint="^xclock$", window_metadata=window_metadata)
        assert window_id

        (position_x, position_y) = (
            getwindowgeometry(window_id=window_id).group("position").split(",")
        )
        delta_x = randint(10, 50)
        delta_y = randint(10, 50)
        position_new = f"{int(position_x) + delta_x},{int(position_y) + delta_y}"

        windowmove(position=position_new, sync=True, window_id=window_id)
        # TODO: Apparently, there is something going on w/ display managers, as this doesn't work ?!?
        #       ... fallback on, did it move "somewhere" ...
        # assert get_window_geometry(window_id=window_id).group("position") == position_new
        assert (
            getwindowgeometry(window_id=window_id).group("position")
            != f"{position_x},{position_y}"
        )
    finally:
        kill_all_xclock_instances()


@pytest.mark.xclock
def test_windowsize():
    """Tests that a window can be sized."""
    try:
        window_metadata = launch_command(args=["xclock"])
        window_id = guess_window(title_hint="^xclock$", window_metadata=window_metadata)
        assert window_id

        (width, height) = (
            getwindowgeometry(window_id=window_id).group("geometry").split("x")
        )
        delta_x = randint(10, 50)
        delta_y = randint(10, 50)
        geometry_new = f"{int(width) + delta_x}x{int(height) + delta_y}"

        windowsize(geometry=geometry_new, sync=True, window_id=window_id)
        assert getwindowgeometry(window_id=window_id).group("geometry") == geometry_new
    finally:
        kill_all_xclock_instances()
