#!/usr/bin/env python

# pylint: disable=redefined-outer-name

"""xdotoolp tests."""

import logging

from shutil import rmtree

import pytest

from _pytest.tmpdir import TempPathFactory

from xdotoolp import (
    get_config_dirs,
    guess_window,
    launch_command,
)

from .testutils import kill_all_xclock_instances, temporary_environment_variable

LOGGER = logging.getLogger(__name__)


def test_get_config_dirs(tmp_path_factory: TempPathFactory):
    """Tests configuration directory retrieval."""
    tmpdir_home = tmp_path_factory.mktemp(f"{__name__}")
    tmpdir_default0 = tmpdir_home.joinpath(".config/xdotoolp")
    tmpdir_default1 = tmpdir_home.joinpath(".xdotoolp")
    tmpdir_xdg_config_home = tmpdir_home.joinpath("xdg_config_home/xdotoolp")
    tmpdir_xdotoolp_configdir = tmpdir_home.joinpath("xdotoolp_configdir")
    for path in [
        tmpdir_default0,
        tmpdir_default1,
        tmpdir_xdg_config_home,
        tmpdir_xdotoolp_configdir,
    ]:
        path.mkdir(parents=True)
    with temporary_environment_variable(
        key="HOME", value=str(tmpdir_home)
    ), temporary_environment_variable(
        key="XDOTOOLP_CONFIGDIR", value=str(tmpdir_xdotoolp_configdir)
    ):
        config_dirs = list(get_config_dirs())
        assert tmpdir_default0 in config_dirs
        assert tmpdir_default1 in config_dirs
        assert tmpdir_xdg_config_home not in config_dirs
        assert tmpdir_xdotoolp_configdir in config_dirs
        with temporary_environment_variable(
            key="XDG_CONFIG_HOME", value=str(tmpdir_xdg_config_home.parent)
        ):
            config_dirs = list(get_config_dirs())
            assert tmpdir_default0 not in config_dirs
            assert tmpdir_default1 in config_dirs
            assert tmpdir_xdg_config_home in config_dirs
            assert tmpdir_xdotoolp_configdir in config_dirs
    rmtree(tmpdir_home, ignore_errors=True)


@pytest.mark.xclock
def test_guess_window():
    """Tests that we can guess for a window (at least sometimes ...)."""
    try:
        window_metadata = launch_command(args=["xclock"])
        window_id = guess_window(title_hint="^xclock$", window_metadata=window_metadata)
        assert window_id
    finally:
        kill_all_xclock_instances()


@pytest.mark.xclock
def test_launch_command():
    """Tests that commands can be launched."""
    try:
        window_metadata = launch_command(args=["xclock"])
        assert window_metadata
        assert [window for window in window_metadata if "xclock" in window.name]
    finally:
        kill_all_xclock_instances()
