#!/usr/bin/env python

"""xdotool wrappers."""

import logging
import re

from .utils import run

EXTENSIONS = ["json", "yaml", "yml"]
LOGGER = logging.getLogger(__name__)
PATTERN_GEOMETRY = pattern = re.compile(
    r"Window\s+(?P<window>\d+)\s+Position:\s(?P<position>\S+)\s\(screen: (?P<screen>\d+)\)\s+Geometry:"
    r"\s(?P<geometry>\S+)"
)
XDOTOOL = "xdotool"


def get_desktop() -> str:
    """Retrieves the current desktop."""
    return run(args=[XDOTOOL, "get_desktop"])


def get_desktop_for_window(*, window_id: str) -> str:
    """Retrieves the desktop containing a given window."""
    return run(args=[XDOTOOL, "get_desktop_for_window", window_id])


def get_num_desktops() -> str:
    """Retrieves the number of desktops."""
    return run(args=[XDOTOOL, "get_num_desktops"])


def getactivewindow() -> str:
    """Retrieves the active window ID."""
    return run(args=[XDOTOOL, "getactivewindow"])


def getwindowgeometry(*, window_id: str) -> re.Match:
    """Retrieved the geometry for a given window."""
    stdout = run(args=[XDOTOOL, "getwindowgeometry", window_id])
    match = PATTERN_GEOMETRY.match(stdout)
    if not match:
        raise Exception(f"Unable to parse output:\n{stdout}")
    return match


def getwindowname(*, window_id: str) -> str:
    """Retrieves the name for a given window."""
    return run(args=[XDOTOOL, "getwindowname", window_id])


def getwindowpid(*, window_id: str) -> str:
    """Retrieves the PID for a given window."""
    return run(args=[XDOTOOL, "getwindowpid", window_id])


def search(*args) -> str:
    """Search for windows."""
    return run(args=[XDOTOOL, "search", *args])


def selectwindow() -> str:
    """Graphically selects window (interactive via UI)."""
    return run(args=[XDOTOOL, "selectwindow"])


def set_desktop(*, desktop: str) -> str:
    """Assigns the current desktop."""
    return run(args=[XDOTOOL, "set_desktop", desktop])


def set_desktop_for_window(*, window_id: str, desktop: str) -> str:
    """Assigns a desktop to a given desktop."""
    LOGGER.debug("Assigning desktop %s to window: %s", desktop, window_id)
    return run(args=[XDOTOOL, "set_desktop_for_window", window_id, str(desktop)])


def windowactivate(*, window_id: str, sync: bool = False) -> str:
    """Activates a given window."""
    LOGGER.debug("Activating window: %s", window_id)
    sync = ["--sync"] if sync else []
    return run(args=[XDOTOOL, "windowactivate", *sync, window_id])


def windowmove(*, window_id: str, position: str, sync: bool = False) -> str:
    """Moves a window to a given position."""
    (position_x, position_y) = position.split(",")
    LOGGER.debug("Moving window %s to: %s", window_id, position)
    sync = ["--sync"] if sync else []
    return run(args=[XDOTOOL, "windowmove", *sync, window_id, position_x, position_y])


def windowsize(*, window_id: str, geometry: str, sync: bool = False) -> str:
    """Resizes a window to a given geometry."""
    (width, height) = geometry.split("x")
    LOGGER.debug("Sizing window %s to: %s", window_id, geometry)
    sync = ["--sync"] if sync else []
    return run(args=[XDOTOOL, "windowsize", *sync, window_id, width, height])
