#!/usr/bin/env python

"""xdotoolp command line interface."""

import logging
import multiprocessing
import os
import re
import subprocess
import sys

from pathlib import Path
from shutil import which
from tempfile import TemporaryDirectory
from textwrap import dedent
from time import sleep
from traceback import print_exception
from typing import Generator, List, NamedTuple, Optional, TypedDict

import click
import yaml

from click.core import Context

from .utils import (
    LOGGING_DEFAULT,
    logging_options,
    run,
    set_log_levels,
)
from .xdotool import (
    get_desktop,
    get_desktop_for_window,
    getwindowgeometry,
    getwindowname,
    getwindowpid,
    search,
    selectwindow,
    set_desktop_for_window,
    windowactivate,
    windowmove,
    windowsize,
)

EXTENSIONS = ["json", "yaml", "yml"]
LOGGER = logging.getLogger(__name__)
XDOTOOL = "xdotool"


class TypingContextObject(TypedDict):
    # pylint: disable=missing-class-docstring
    verbosity: int


class TypingWindowMetadata(NamedTuple):
    # pylint: disable=missing-class-docstring
    id: str
    name: str


def do_load(*, path: Path, verbosity: int):
    """
    Loads a given xdotoolp configuration file.

    Args:
        path: Path of the configuration file to be loaded.
        verbosity: Verbosity level under which to execute.
    """
    try:
        LOGGER.info("Loading: %s", path)
        config = yaml.load(path.read_text(encoding="utf-8"), Loader=yaml.SafeLoader)
        # LOGGER.debug("Configuration:\n%s", yaml.dump(config))

        desktop_default = config.get("default_desktop", 0)
        for window in config["windows"]:
            # Construct the child environment ...
            do_not_copy_environment = bool(window.get("do_not_copy_environment", False))
            env = {}
            if not do_not_copy_environment:
                env = os.environ.copy()
            if "environment" in window:
                env.update(window["environment"])

            # Start the process, find the window ...
            LOGGER.debug("Executing: %s", window["command"])
            with open(os.devnull, "wb") as devnull:
                window_metadata = launch_command(
                    args=window["command"],
                    cwd=window.get("start_directory", "/"),
                    env=env,
                    preexec_fn=os.setpgrp(),
                    shell=bool(window.get("shell", False)),
                    stderr=devnull,
                    stdout=devnull,
                )
            window["id"] = guess_window(
                title_hint=window.get("title-hint"), window_metadata=window_metadata
            )
            LOGGER.debug("Guessed window ID: %s", window["id"])
            if not window["id"]:
                LOGGER.error("Unable to locate spawned window!")
                continue

            # Position the window ...
            set_desktop_for_window(
                desktop=window.get("desktop", desktop_default), window_id=window["id"]
            )
            if "geometry" in window:
                windowsize(
                    geometry=window["geometry"], sync=True, window_id=window["id"]
                )
            if "position" in window:
                windowmove(
                    position=window["position"], sync=True, window_id=window["id"]
                )

        # Activate a (single) window, after all windows are finished being placed ...
        windows = [
            window
            for window in config["windows"]
            if "focus" in window and bool(window["focus"])
        ]
        if len(windows) > 1:
            LOGGER.error(
                "Only 1 window may defined as focusable; refusing to set focus!"
            )
        elif len(windows) > 0:
            windowactivate(sync=True, window_id=windows[0]["id"])

    except Exception as exception:  # pylint: disable=broad-except
        if verbosity > 0:
            logging.fatal(exception)
        if verbosity > LOGGING_DEFAULT:
            exc_info = sys.exc_info()
            print_exception(*exc_info)
        sys.exit(1)


def get_config_dirs() -> Generator[Path, None, None]:
    """Returns the xdotoolp configuration directory(ies)."""
    paths = []
    if "XDOTOOLP_CONFIGDIR" in os.environ:
        paths.append(os.environ["XDOTOOLP_CONFIGDIR"])
    if "XDG_CONFIG_HOME" in os.environ:
        paths.append(os.path.join(os.environ["XDG_CONFIG_HOME"], "xdotoolp"))
    else:
        paths.append("~/.config/xdotoolp")
    paths.append("~/.xdotoolp")
    paths = [Path(path).expanduser() for path in paths]

    for path in paths:
        if path.exists():
            yield path


def get_context_object(*, context: Context) -> TypingContextObject:
    """Wrapper method to enforce type checking."""
    return context.obj


def guess_window(
    *, title_hint: str = None, window_metadata: List[TypingWindowMetadata]
) -> Optional[str]:
    """Attempts to isolate a single window from a list of potential matches."""
    LOGGER.debug(
        "Guessing against %d windows using title-hint: %s",
        len(window_metadata),
        str(title_hint),
    )
    if len(window_metadata) < 1:
        return None

    # Quick an dirty ...
    if len(window_metadata) == 1:
        return window_metadata[0].id

    # If we have hint, use it; otherwise, try looking for things with ANY title ...
    pattern = re.compile(title_hint if title_hint is not None else r"^.+$")
    guesses = [window for window in window_metadata if pattern.match(window.name)]
    if len(guesses) == 1:
        LOGGER.debug("Found matching title: %s", guesses[0].name)
        return guesses[0].id

    if len(guesses) > 1:
        LOGGER.warning(
            "Too many matching titles; try %s 'title-hint'!",
            "constraining" if title_hint else "adding",
        )
        # ... it should still be better to use things with titles ...
        window_metadata = guesses
    else:
        LOGGER.warning(
            "Too few matching titles; try %s 'title-hint'!",
            "relaxing" if title_hint else "adding",
        )

    LOGGER.debug("Best effort at an id-based match ...")
    # The greater the id, the later the window was created?!? ¯\_(ツ)_/¯
    return sorted(window_metadata, key=lambda x: int(x.id), reverse=True)[0].id


def launch_command(
    *, delay: int = 1, tries: int = 3, **kwargs
) -> List[TypingWindowMetadata]:
    """Executes a command and attempts to identify the window(s) that were created as a result."""

    search_args = [
        "--class",
        "--classname",
        "--maxdepth=2",
        "--name",
        "--onlyvisible",
        ".*",
    ]
    windows_ids_before = search(*search_args).split("\n")

    def child():
        with subprocess.Popen(**kwargs) as process:
            LOGGER.debug("Started pid: %s", process.pid)

    context = multiprocessing.Process(name="child", target=child)
    context.start()

    result = []
    for _ in range(tries):
        windows_ids_after = search(*search_args).split("\n")

        delta = [x for x in windows_ids_after if x not in windows_ids_before]
        if delta:
            for window_id in delta:
                name = getwindowname(window_id=window_id)
                result.append(TypingWindowMetadata(id=window_id, name=name))
            break
        sleep(delay)

    return result


@click.group()
@logging_options
@click.pass_context
def cli(
    context: Context,
    verbosity: int = LOGGING_DEFAULT,
):
    """A declarative window instantiation utility based on xdotool."""

    if verbosity is None:
        verbosity = LOGGING_DEFAULT

    set_log_levels(verbosity)

    context.obj = {
        "verbosity": verbosity,
    }


@cli.command()
def learn():
    """
    Capture metadata from a graphically selected window.

    Once execute, the cursor of the display manger should be altered until a window is selected by clicking on it.
    """
    window_id = selectwindow()
    LOGGER.debug("Window: %s", window_id)

    pid = getwindowpid(window_id=window_id)
    command = run(args=["ps", "-o", "cmd", "-p", pid, "h"])

    desktop = get_desktop_for_window(window_id=window_id)
    match = getwindowgeometry(window_id=window_id)
    template = dedent(
        f"""
        windows:
        - command: {command}
          desktop: {desktop}
          geometry: {match.group("geometry")}
          position: {match.group("position")}
        """
    )
    LOGGER.info(template)


@cli.command()
@click.argument("config", nargs=-1)
@click.pass_context
def load(context: Context, config):
    """Load configuration file(s)."""
    ctx = get_context_object(context=context)

    configs = []
    paths = [Path(c) for c in config]
    for path in paths:
        # Is it qualified or relative to the CWD?
        if path.exists():
            configs.append(path)
            continue

        # Is it relative to a configuration directory?
        for config_dir in get_config_dirs():
            lpath = config_dir.joinpath(path)
            if lpath.exists():
                configs.append(lpath)
                break
            found = False
            for extension in EXTENSIONS:
                lpath = config_dir.joinpath(f"{str(path)}.{extension}")
                if lpath.exists():
                    configs.append(lpath)
                    found = True
                    break
            if found:
                break

    for path in configs:
        do_load(path=path, verbosity=ctx["verbosity"])


@cli.command()
def ls():
    # pylint: disable=invalid-name
    """List configuration file(s)."""
    files = []
    for config_dir in get_config_dirs():
        for extension in EXTENSIONS:
            files.extend(config_dir.glob(f"**/*.{extension}"))
    for file in sorted(files):
        print(file)


@cli.command()
@click.pass_context
def test(context: Context):
    """Perform basic acceptance tests."""
    ctx = get_context_object(context=context)

    LOGGER.info("Python Version:\n\t%s\n", "\n\t".join(sys.version.split("\n")))
    LOGGER.info(
        "Configuration Directories:\n\t%s\n",
        "\n\t".join([str(path) for path in get_config_dirs()]),
    )
    LOGGER.info("Tool Location:\n\t%s\n", which(XDOTOOL) or f"'{XDOTOOL}' not found!")
    LOGGER.info("Subprocess Test:\n\t%s\n", run(args="pwd"))

    with TemporaryDirectory() as tmpdir:
        path = Path(tmpdir).joinpath("xclock.yml")
        data = dedent(
            f"""
            default_desktop: {get_desktop()}
            windows:
            - command: xclock
              geometry: 300x300
              position: 25,25
              shell: true
              title-hint: ^xclock$
            - command:
              - xclock
              - -digital
              geometry: 300x40
              position: 25,375
              title-hint: ^xclock$
            """
        )
        path.write_text(data=data, encoding="utf-8")
        LOGGER.info("Test Configuration:\n\t%s\n", "\n\t".join(data.split("\n")))
        do_load(path=path, verbosity=ctx["verbosity"])


@cli.command()
def version():
    """Displays the utility version."""
    # Note: This cannot be imported above, as it causes a circular import!
    from . import __version__  # pylint: disable=import-outside-toplevel

    print(__version__)


if __name__ == "__main__":
    # pylint: disable=no-value-for-parameter
    cli()
